#pragma once
#include "MatBaza.h"
#include "Matrica4x4.h"
//#include "Slika.h"

class Slika;
class SlikaGPU;

/** Osnovna virtualna kamera za renderiranje.
 *
 *  Ova klasa podržava promjene FOV-a, žarišne duljine i općenitih transformacija pogleda virtualne kamere u 3D koordinatnom sustavu. U ovom konkretnom programu, koristi se samo radi raytracinga, ali mogla bi biti iskorištena za rasterizaciju.
 *
 *  Svaka kamera može koristiti ili perpsketivnu ili ortografsku projekciju, što se može odabrati prilikom konstrukcije ili kasnije slobodno izmijeniti, što utječe ili na konačne transformacije u screenspace (rasterizacija) ili jednostavno na način računanja izlaznih zraka (raytracing).
 */
class Kamera
{
  friend Slika;
  friend SlikaGPU;

 public:
  /** Tip projekcije virtualne kamere.
   */
  enum class TipProjekcije//ovo nisu kose projekcije
  {
    Perspektivna, /**< Perspektivni tip projekcije (za realistične prikaze) */
      Ortografska   /**< Ortografska projekcija (sve zrake su paralelne) */
  };

  //  Matrica4x4 transformacija;

  /** Glavni konstruktor za virtualnu kameru. Uzima sve potrebne inicijalne projektivne parametre.
   * \param[in] proj  Tip projekcije.
   * \param[in] o     Početna pozicija (ishodište viewspace-a) kamere.
   * \param[in] d,g,i Ortonormirani vektori desno orijentiranog virtualnog okvira kamere (desno, gore, iza).
   * \param[in] fov   Horizontalni kut pogleda (FOV) kamere u radijanima.
   * \param[in] w,h   Dimenzije slike (trenutno je nužno da budu identične piksel rezoluciji).
   */
  __device__ __host__ Kamera(TipProjekcije proj,const Vektor4 &o,const Vektor3 &d,const Vektor3 &g,const Vektor3 &i,Realan fov,int sirina,int visina);
  __device__ __host__ Kamera() = default;
  __device__ __host__ ~Kamera() = default;

  __device__ __host__ Kamera(const Kamera &) = default;
  __device__ __host__ Kamera & operator=(const Kamera &) = default;
  __device__ __host__ Kamera(Kamera &&) = default;
  __device__ __host__ Kamera & operator=(Kamera &&) = default;

  /** Dinamički mijenja žarišnu duljinu kamere na zadani realan broj.
   * \param[in] d Nova žarišna duljina u viewspace jedinicama.
   */
  __device__ __host__ void PostaviDuljinu(Realan d);

  /** Dinamički mijenja horizontalni kut pogleda kamere na zadani realan broj.
   * \param[in] fov Novi horizontalni kut pogleda u radijanima.
   */
  __device__ __host__ void PostaviFOV(Realan fov);

  /** Dinamički mijenja dimenzije slike na zadani par cijelih brojeva.
   * \param[in] sirina,visina Nove dimenzije u ekranskim pikselima.
   */
  __device__ __host__ void PostaviDimenzije(int sirina, int visina);

  /** Postavlja apsolutnu 4x4 transformaciju na kameru.
   *
   *  Ova metoda *neće* komponirati transformacije; dakle, sve se provodi nad inicijalnim okvirom i pozicijom kamere!
   * \param[in] tran 4x4 transformacijska matrica.
   */
  __device__ __host__ void PostaviTransformaciju(const Matrica4x4 &tran);

  /** Komponira 4x4 transformaciju na kameru.
   *
   *  Ova metoda komponira zadanu transformaciju s trenutnom modelview matricom, što rezultira u kumulativnom efektu transformacije, korisnom za animacije kamere i sl.
   * \param[in] tran 4x4 transformacijska matrica.
   */
  __device__ __host__ void KomponirajTransformaciju(const Matrica4x4 &tran);

  /** Dinamički mijenja tip projekcije kamere.
   * \param[in] tip Novi tip projekcije kamere.
   */
  __device__ __host__ void PostaviTip(TipProjekcije tip) {projekcija = tip;}

  /** Usmjerava kameru u fiksnu točku u svjetskim koordinatama.
   * \param[in] cilj  Točka u svjetskom prostoru u koju kamera mora gledati.
   * \param[in] gore  Normiran vektor usmjeren ravno iznad kamere, dakle ekvivalentan Y osi u viewspace-u.
   */
  __device__ __host__ void GledajU(const Vektor4 &cilj,const Vektor3 &gore);

  /** Vraća trenutnu poziciju kamere u svjetskom prostoru.
   * \return 3-vektor trenutne pozicije kamere u svjetskim koordinatama.
   */
  __device__ __host__ Vektor3 Pozicija() const;

  /** Vraća trenutni vektor "desno" okvira kamere, u svjetskim koordinatama.
   * \return Normiran vektor u svjetskim koordinatama koji u viewspace-u predstavlja X os.
   */
  __device__ __host__ const Vektor3 &U() const;

  /** Vraća trenutni vektor "gore" okvira kamere, u svjetskim koordinatama.
   * \return Normiran vektor u svjetskim koordinatama koji u viewspace-u predstavlja Y os.
   */
  __device__ __host__ const Vektor3 &V() const;

  /** Vraća trenutni vektor "iza" okvira kamere, u svjetskim koordinatama.
   * \return Normiran vektor u svjetskim koordinatama koji u viewspace-u predstavlja Z os.
   */
  __device__ __host__ const Vektor3 &W() const;

  /** Vraća trenutnu žarišnu duljinu kamere.
   * \return Žarišna duljina kamere u viewspace jedinicama.
   */
  __device__ __host__ Realan Duljina() const;

  /** Vraća trenutnu X dimenziju slike kamere.
   * \return X dimenzija (širina) slike.
   */
  __device__ __host__ int Sirina() const;

  /** Vraća trenutnu Y dimenziju slike kamere.
   * \return Y dimenzija (visina) slike.
   */
  __device__ __host__ int Visina() const;

  /** Vraća trenutni horizontalni kut pogleda (HFOV) kamere.
   * \return HFOV kamere u radijanima.
   */
  __device__ __host__ Realan HFOV() const {return hfov;}

  /** Vraća trenutni tip projekcije kamere.
   * \return Tip projekcije kamere.
   */
  __device__ __host__ TipProjekcije Tip() const {return projekcija;}

 private:
  /** Pomoćna privatna funkcija.
   *
   *  Pomoćna funkcija koja postavlja okvir odnosno granice slikovne ravnine kamere pri postavljanju i/ili promjeni projektivnih parametara kamere.
   */
  __device__ __host__ void PostaviGranice();

  /** Trenutni tip projekcije kamere.
   */
  TipProjekcije projekcija;
  //  Matrica4x4 transformacija;

  /** Inicijalni okvir kamere.
   */
  Vektor3 x, y, z; //inicijalni okvir
  /** Trenutni okvir kamere.
   */
  Vektor3 u, v, w;

  /** Inicijalna pozicija (ishodište) kamere.
   */
  Vektor4 pozicija0;
  /** Trenutna pozicija kamere.
   */
  Vektor4 pozicija;

  /** Trenutni horizontalni kut pogleda kamere.
   */
  Realan hfov;//horizontalni kut pogleda (FOV)
  /** Trenutna žarišna duljina kamere.
   */
  Realan d;//žarišna duljina
  /** Trenutne granice slike kamere.
   */
  Realan l, r, b, t;
  /** Trenutne apsolutne piksel dimenzije slike kamere.
   */
  int sirina, visina;//razlučivost
};
