//template...?
#pragma once
#include "MatBaza.h"
#include "Vektor3.h"

/** Tip općenite 3x3 matrice s odgovarajućim operacijama.
 *
 *  Osnovni tip matrice za predstavljanje bilo kojeg linearnog operatora u trodimenzionalnom vektorskom prostoru.
 */
class Matrica3x3
{
  //potrebno za komutacijsko svojstvo
  /** Množenje matrica reda 3.
   *  \param[in] matA  Lijevi matrični operand.
   *  \param[in] matB  Desni matrični operand.
   *  \return 3x3 matrica produkta.
   */
  friend __device__ __host__ Matrica3x3 operator*(const Matrica3x3 &matA, const Matrica3x3 &matB);
  /** Skalar-matrično množenje s matricama reda 3.
   *  \param[in] skalar Lijevi skalarni operand.
   *  \param[in] mat    Desni matrični operand.
   *  \return 3x3 rezultantna matrica.
   */
  friend __device__ __host__ Matrica3x3 operator*(Realan skalar, const Matrica3x3 &mat);
  /** Skalar-matrično množenje s matricama reda 3.
   *  \param[in] skalar Desni skalarni operand.
   *  \param[in] mat    Lijevi matrični operand.
   *  \return 3x3 rezultantna matrica.
   */ 
  friend __device__ __host__ Matrica3x3 operator*(const Matrica3x3 &mat, Realan skalar);
  /** Matrično-vektorsko množenje reda 3.
   *  \param[in] mat Lijevi matrični operand.
   *  \param[in] vek Desni vektorski operand.
   *  \return Rezultantni 3-vektor.
   */
  friend __device__ __host__ Vektor3 operator*(const Matrica3x3 &mat, const Vektor3 &vek);

 private:
  static constexpr Realan epsilon = 1.0e-8f;//usporedni epsilon
  bool determinantaSpremna = false;
  //bool invertiranaSpremna = false;
  //  Matrica4x4 inverz;
  /** Determinanta matrice.
   */
  Realan determinanta;
  /** Interno 2D polje u kojem su pohranjeni elementi matrice.
   */
  Realan m[3][3];

  Realan m1, m2, m3; //minore izračunate pri računu determinante

 public:
  /** Default konstruktor stvara nulmatricu.
   */
  __device__ __host__ Matrica3x3();
  __device__ __host__ ~Matrica3x3() = default;
  /** Konstruktor matrice element-po-element.
   *
   *  Elementi se navode slijeva na desno, od najvišeg do najnižeg retka.
   */
  __device__ __host__ Matrica3x3(Realan,Realan,Realan,Realan,Realan,Realan,Realan,Realan,Realan);
  /** Stupčani konstruktor matrice.
   *  \param[in] stupci Polje od triju stupaca, redom slijeva u desno.
   */
  __device__ __host__ Matrica3x3(const Vektor3 stupci[3]);
  /** Konstruktor matrice od 2D polja.
   *
   *  \param[in] polje 2D polje koje prirodno odgovara matrici (row-major).
   */
  __device__ __host__ explicit Matrica3x3(const Realan polje[3][3]);
  __device__ __host__ Matrica3x3(const Matrica3x3 &) = default;
  __device__ __host__ Matrica3x3 & operator=(const Matrica3x3 &) = default;

  __device__ __host__ Matrica3x3(Matrica3x3 &&) = default;
  __device__ __host__ Matrica3x3 & operator=(Matrica3x3 &&) = default;

  /** Matrično zbrajanje reda 3.
   *  \param[in] druga Desni matrični operand zbrajanja.
   *  \return Suma trenutne (lijeve) matrice i dane desne matrice reda 3.
   */
  __device__ __host__ Matrica3x3 operator+(const Matrica3x3 &druga) const;
  /** Matrično oduzimanje reda 3.
   *  \param[in] druga Desni matrični operand oduzimanja.
   *  \return Razlika trenutne (lijeve) matrice i dane desne matrice reda 3.
   */
  __device__ __host__ Matrica3x3 operator-(const Matrica3x3 &druga) const;
  /** Negiranje matrice reda 3 (unarni operator).
   */
  __device__ __host__ void operator-();//unarni

  /** Vraća traženi element (i,j) matrice.
   *
   *  Indeksi počinju od 0.
   *  \param[in] i Indeks retka.
   *  \param[in] j Indeks stupca.
   *  \return Element (i,j)
   */
  __device__ __host__ Realan Element(int i,int j) const;

  /** Provjera jednakosti matrica reda 3.
   *  \param[in] druga Druga matrica za usporedbu.
   *  \return `true` ako su matrice jednake; inače `false`.
   */
  __device__ __host__ bool operator==(const Matrica3x3 &druga) const;
  /** Provjera nejednakosti matrica reda 3.
   *  \param[in] druga Druga matrica za usporedbu.
   *  \return `true` ako matrice nisu jednake; inače `false`.
   */
  __device__ __host__ bool operator!=(const Matrica3x3 &druga) const;

  /** Transponiranje matrice reda 3.
   *  \return Transponirana matrica.
   */
  __device__ __host__ Matrica3x3 Transponiraj() const;
  /** Invertiranje matrice reda 3.
   *  \return Inverz ove matrice.
   */
  __device__ __host__ Matrica3x3 Invertiraj();
  //  double Minora(int,int) const;
  /** Vraća determinantu matrice reda 3.
   *  \return Determinanta ove matrice.
   */
  __device__ __host__ Realan Determinanta();

  /** Vraća jediničnu matricu reda 3.
   */
  static __device__ __host__ Matrica3x3 Jedinicna();

  /** Vraća rotacijsku matricu po osi X.
   *  \param[in] radijana Kut u radijanama za rotaciju oko osi X.
   */
  static __device__ __host__ Matrica3x3 RotirajX(Realan radijana);
  /** Vraća rotacijsku matricu po osi Y.
   *  \param[in] radijana Kut u radijanama za rotaciju oko osi Y.
   */
  static __device__ __host__ Matrica3x3 RotirajY(Realan radijana);
  /** Vraća rotacijsku matricu po osi Z.
   *  \param[in] radijana Kut u radijanama za rotaciju oko osi Z.
   */
  static __device__ __host__ Matrica3x3 RotirajZ(Realan radijana);
};
