//template?
#pragma once
#include "MatBaza.h"
#include "Matrica3x3.h"
#include "Vektor4.h"
#include "Vektor3.h"

/** Tip općenite 4x4 matrice s odgovarajućim operacijama.
 *
 *  Osnovni tip matrice za predstavljanje bilo kojeg afinog operatora (moglo bi se proširiti i na druge projektivne transformacije) u trodimenzionalnom afinom prostoru.
 */
class Matrica4x4
{
  //potrebno za komutacijsko svojstvo
  /** Množenje matrica reda 4.
   *  \param[in] matA  Lijevi matrični operand.
   *  \param[in] matB  Desni matrični operand.
   *  \return 4x4 matrica produkta.
   */
  friend __device__ __host__ Matrica4x4 operator*(const Matrica4x4 &matA, const Matrica4x4 &matB);
  /** Skalar-matrično množenje s matricama reda 4.
   *  \param[in] skalar Lijevi skalarni operand.
   *  \param[in] mat    Desni matrični operand.
   *  \return 4x4 rezultantna matrica.
   */
  friend __device__ __host__ Matrica4x4 operator*(Realan skalar, const Matrica4x4 &mat);
  /** Skalar-matrično množenje s matricama reda 4.
   *  \param[in] skalar Desni skalarni operand.
   *  \param[in] mat    Lijevi matrični operand.
   *  \return 4x4 rezultantna matrica.
   */
  friend __device__ __host__ Matrica4x4 operator*(const Matrica4x4 &mat, Realan skalar);
  /** Matrično-vektorsko množenje reda 4.
   *  \param[in] mat Lijevi matrični operand.
   *  \param[in] vek Desni vektorski operand.
   *  \return Rezultantni 4-vektor.
   */
  friend __device__ __host__ Vektor4 operator*(const Matrica4x4 &mat, const Vektor4 &vek);

 private:
  static constexpr Realan epsilon = 1.0e-8f;//usporedni epsilon
  bool determinantaSpremna = false;

  /** Determinanta matrice.
   */
  Realan determinanta;
  /** Interno 2D polje u kojem su pohranjeni elementi matrice.
   */
  Realan m[4][4];

  Realan t1, t2, t3, t4, t5, t6;//minore 2. reda
  Realan m1, m2, m3, m4; //minore izračunate pri računu determinante

 public:
  /** Default konstruktor stvara nulmatricu.
   */
  __device__ __host__ Matrica4x4();
  __device__ __host__ ~Matrica4x4() = default;
  /** Konstruktor matrice element-po-element.
   *
   *  Elementi se navode slijeva na desno, od najvišeg do najnižeg retka.
   */
  __device__ __host__ Matrica4x4(Realan,Realan,Realan,Realan,Realan,Realan,Realan,Realan,Realan,Realan,Realan,Realan,Realan,Realan,Realan,Realan);
  /** Konstruktor matrice od 2D polja.
   *
   *  \param[in] polje 2D polje koje prirodno odgovara matrici (row-major).
   */
  explicit __device__ __host__ Matrica4x4(const Realan[4][4]);
  __device__ __host__ Matrica4x4(const Matrica4x4 &) = default;
  __device__ __host__ Matrica4x4 & operator=(const Matrica4x4 &) = default;

  __device__ __host__  Matrica4x4(Matrica4x4 &&) = default;
  __device__ __host__ Matrica4x4 & operator=(Matrica4x4 &&) = default;

  /** Matrično zbrajanje reda 4.
   *  \param[in] druga Desni matrični operand zbrajanja.
   *  \return Suma trenutne (lijeve) matrice i dane desne matrice reda 4.
   */
  __device__ __host__ Matrica4x4 operator+(const Matrica4x4 &) const;
  /** Matrično oduzimanje reda 4.
   *  \param[in] druga Desni matrični operand oduzimanja.
   *  \return Razlika trenutne (lijeve) matrice i dane desne matrice reda 4.
   */
  __device__ __host__ Matrica4x4 operator-(const Matrica4x4 &) const;
  /** Negiranje matrice reda 4 (unarni operator).
   */
  __device__ __host__ void operator-();//unarni

  /** Vraća traženi element (i,j) matrice.
   *
   *  Indeksi počinju od 0.
   *  \param[in] i Indeks retka.
   *  \param[in] j Indeks stupca.
   *  \return Element (i,j)
   */
  __device__ __host__ Realan Element(int,int) const;

  /** Provjera jednakosti matrica reda 4.
   *  \param[in] druga Druga matrica za usporedbu.
   *  \return `true` ako su matrice jednake; inače `false`.
   */
  __device__ __host__ bool operator==(const Matrica4x4 &) const;

  /** Provjera nejednakosti matrica reda 4.
   *  \param[in] druga Druga matrica za usporedbu.
   *  \return `true` ako matrice nisu jednake; inače `false`.
   */
  __device__ __host__ bool operator!=(const Matrica4x4 &) const;

  /** Transponiranje matrice reda 4.
   *  \return Transponirana matrica.
   */
  __device__ __host__ Matrica4x4 Transponiraj() const;
  /** Invertiranje matrice reda 4.
   *  \return Inverz ove matrice.
   */
  __device__ __host__ Matrica4x4 Invertiraj();

  /** Vraća determinantu matrice reda 4.
   *  \return Determinanta ove matrice.
   */
  __device__ __host__ Realan Determinanta();

  /** Vraća odgovarajući linearnu transformaciju.
   *
   *  Vraća podmatricu reda 3 koja predstavlja linearni operator jedinstveno dobiven zanemarivanjem translacije trenutne matrice.
   *  \return Matrica reda 3 odgovarajućeg linearnog operatora.
   */
  __device__ __host__ Matrica3x3 Linearna() const;

  /** Vraća jediničnu matricu reda 4.
   */
  static __device__ __host__ Matrica4x4 Jedinicna();

  /** Vraća transformaciju koja uniformno skalira po svim osima.
   *  \param[in] uniformSkala Skalar za uniformno skaliranje.
   */
  static __device__ __host__ Matrica4x4 Skaliraj(Realan uniformSkala);

  static __device__ __host__ Matrica4x4 RotirajX(Realan radijana);
  static __device__ __host__ Matrica4x4 RotirajY(Realan radijana);
  static __device__ __host__ Matrica4x4 RotirajZ(Realan radijana);

  /** Vraća translacijsku matricu.
   *  \param[in] vektor Vektor pomaka za translaciju.
   */
  static __device__ __host__ Matrica4x4 Translatiraj(const Vektor3 &vektor);
};
