#pragma once
#include "ObjektGPU.h"
#include "TrokutGPU.h"
#include "SferaGPU.h"
#include "TrokutModelGPU.h"
#include "Scena.h"
#include <memory>

/** Struktura za podatke o GPU sceni.
 *
 *  U ovoj strukturi su sve informacije potrebne za predstaviti scenu GPU primitiva. Bilo kakva lista se predstavlja poljem koje host dinamički alocira u GPU globalnoj memoriji. Dakle, svi pokazivači ovdje pokazuju u globalnu GPU memoriju!
 */
struct ScenaPodaci
{
  int brojSvjetala = 0;
  int brojSfera = 0;
  int brojTrokutModela = 0;
  int brojTrokuta = 0;
  
  Svjetlo *svjetla = nullptr; //GPU-u šaljemo serijalizirana polja svih potrebnih scenskih podataka jer u HIP-u trenutno nije moguće malloc()-ati stvari na GPU-u (dakle nema STL kontejnera i sl.)
  SferaGPU *sfere = nullptr;
  TrokutModelBaza::Podaci *modeli = nullptr;
  TrokutGPU *trokuti = nullptr;
};


/** Klasa za scenu na GPU-u.
 *
 *  Za sada se scena drukčije reprezentira na GPU-u od CPU reprezentacije. Osnovna ideja je ista kao u klasi Scena.
 */
class ScenaGPU //TODO: ovi podaci bi se svi trebali alocirati
//na GPU-u i ne bi uopće trebali biti dostupni hostu direktno
//za korištenje: kad se instancira ovaj objekt (s overloadanim
//alokatorima hip*) preko tvornice, može se proslijediti samo
//pointer GPU-u budući je sve tamo pohranjeno i ne moramo kopirati
//ScenaGPU objekt svaki put pri dizanju kernela (mada je riječ o
//relativno malom broju podataka)
{
 private:
  /** Svi podaci o sceni.
   */
  ScenaPodaci podaci;
  /** Ukupan broj transformacijskih niti.
   *
   *  Ovo je jednako sumi sljedećih triju članova. Predstavlja ukupan broj niti koje će GPU trebati vrtiti za transformirati sve objekte na sceni.
   */
  size_t brojNiti;
  size_t brojSferaNiti;
  size_t brojTrokutNiti;
  size_t brojModelNiti;

  /** Oslobađa svu GPU memoriju za scenu.
   */
  __host__ void Unisti();

 public:
  __host__ ScenaGPU() = default;
  __host__ ScenaGPU(Scena &);
  __host__ ScenaGPU(const ScenaGPU &) = default; //shallow kopija; ovo se radi samo za prijenos scene u GPU memoriju, kad su već arrayevi alocirani u samoj globalnoj memoriji i ne smije se koristiti za ikakvo drukčije kopiranje na hostu!
  __host__ ScenaGPU &operator=(const ScenaGPU &&);
  __host__ ~ScenaGPU();

  static __device__ __host__ void TransformirajObjekte(ScenaPodaci &podaci);

  // __device__ __host__ void TransformirajObjekte();
  __device__ __host__ SferaGPU *DajSfere() const;
  __device__ __host__ TrokutModelBaza::Podaci *DajModele() const;
  __device__ __host__ TrokutGPU *DajTrokute() const;
  __device__ __host__ Svjetlo *DajSvjetla() const;

  __host__ ScenaPodaci &DajPodatke() {return podaci;}
  
  __device__ __host__ int DajBrojSfera() const;
  __device__ __host__ int DajBrojModela() const;
  __device__ __host__ int DajBrojTrokuta() const;
  __device__ __host__ int DajBrojSvjetala() const;

  __device__ __host__ size_t DajBrojTransformNiti() const;
  __device__ __host__ size_t DajBrojSferaTransformNiti() const {return brojSferaNiti;}
  __device__ __host__ size_t DajBrojTrokutTransformNiti() const {return brojTrokutNiti;}
  __device__ __host__ size_t DajBrojModelTransformNiti() const {return brojModelNiti;}
  
  /*__device__ __host__ int DajBrojTransformNiti() const;
  __device__ __host__ const std::forward_list<Svjetlo> &DajSvjetla() const;
  __device__ __host__ const std::forward_list<Objekt> &DajListu() const;

  __device__ __host__ void DodajSvjetlo(Svjetlo);
  __device__ __host__ void DodajObjekt(Objekt);
  */
};
