#include "SlikaGPU.h"
#include "MatBaza.h"
#include <utility>

__host__ SlikaGPU::SlikaGPU(int sirina, int visina) : sirina(sirina), visina(visina)
{
  hipMalloc(&slika, sizeof(Uint32) * sirina * visina);
  brojBlokX = (sirina + BLOK_X - 1) / BLOK_X;
  brojBlokY = (visina + BLOK_Y - 1) / BLOK_Y;
}

//__device__ __host__ Uint32 *&SlikaGPU::DajSliku() const {return slika;}
__host__ void SlikaGPU::DajSliku(Uint32 *dest) const
{
  hipMemcpy(dest, slika, sizeof(Uint32) * sirina * visina, hipMemcpyDeviceToHost);
  //TODO:ovdje ide kopiranje samo rezerviranog dijela slike koji predstavlja pravu sliku bez paddinga, jer to jedino treba hostu
}

__host__ SlikaGPU::~SlikaGPU()
{
  if(slika)
    hipFree(slika);
}

__device__ __host__ std::pair<int, int> SlikaGPU::DajDimenzije() const
{
  return std::make_pair(sirina, visina);
}

__device__ __host__ std::pair<int, int> SlikaGPU::DajBrojBlokova() const
{
  return std::make_pair(brojBlokX, brojBlokY);
}

__host__ void SlikaGPU::rayCPU(ScenaPodaci scena, Kamera kamera, Uint32 *slika)
{
  for(int i = 0; i < kamera.sirina; ++i)
    for(int j = 0; j < kamera.visina; ++j)
      {

  if(i >= kamera.sirina || j >= kamera.visina)
    return; //izašli smo van ekrana; odustani od rendera za ovaj piksel!
  //probaj optimizaciju: ovo možda dovodi do prevelike divergencije warpova,
  //pa bi bilo bolje alocirati veću padding rezoluciju i nju renderirati
  //na GPU-u, bez obzira što CPU neće iskoristiti cijelu tu površinu kad
  //pročita sliku (orig.rezolucija je ovdje i ona se vraća sa getterima
  //, dok interno u privatnim varijablama GPU cacheira virtualnu rez.!)
  
  Realan u = kamera.l + i + 0.5;
  Realan v = kamera.b + j + 0.5;

  Vektor3 smjer, o;
  switch(kamera.projekcija)//optimizacija: 2 vrste kernela, za 2 vrste kamere jer jedna ne može odjednom postati druga dinamički i ovdje je to suvišno granati!
    {
    case Kamera::TipProjekcije::Ortografska:
      o = kamera.pozicija.Radijusvektor() + u * kamera.u + v * kamera.v;
      smjer = -kamera.w;
      break;

    case Kamera::TipProjekcije::Perspektivna:
      o = kamera.pozicija.Radijusvektor();
      smjer = -kamera.d * kamera.w + u * kamera.u + v * kamera.v;
      break;
    }

  smjer = smjer.Normiraj();
  Zraka zraka(o, smjer);
  SferaGPU *prvaSfera = nullptr;
  const TrokutGPU *prviTrokut = nullptr;
  Realan najmanji = BESKONACNO;
  Vektor3 sjeciste, potencijalnoSjeciste;
  int presjecenTrokut;

  for(int k = 0; k < scena.brojTrokuta; ++k)
    {
      const auto &objekt = scena.trokuti[k];
      Realan t;
      if (!(objekt.PresijeciZraku(zraka, t, potencijalnoSjeciste)))
	continue;
      
      if (t < najmanji && t > 0)//iza svih i ispred kamere
	{
	  najmanji = t;
	  prviTrokut = &objekt;
	  prvaSfera = nullptr;
	  sjeciste = potencijalnoSjeciste;
	}
    }
  
  for(int k = 0; k < scena.brojSfera; ++k) //TODO:prebaci ovo generično računanje u template funkciju (izgubismo polimorfizam)
    {
      auto &objekt = scena.sfere[k];
      Realan t;
      if (!(objekt.PresijeciZraku(zraka, t, potencijalnoSjeciste)))
	continue;

      if (t < najmanji && t > 0)//iza svih i ispred kamere
	{
	  najmanji = t;
	  prvaSfera = &objekt;
	  prviTrokut = nullptr;
	  sjeciste = potencijalnoSjeciste;
	}
    }

    
    Uint32 piksel;//RGB piksel;
    if(!prviTrokut  && !prvaSfera)
      piksel = pozadinskaBoja;

    else
      {
	RGB rgb;

	//piksel = rgb;
	piksel = RGB{255u, 255u, 255u};
      }

    Realan tmpx = static_cast<Realan>(i)/kamera.sirina;
    Realan tmpy = static_cast<Realan>(j)/kamera.visina;
    //slika[i + kamera.sirina * j] = RGB{tmpx, tmpx, tmpx} * RGB{tmpy, 0.0f, 0.0f};
    Realan tmp0 = (smjer[0]+1)/2.0f;
    Realan tmp1 = (smjer[1]+1)/2.0f;
    Realan tmp2 = (smjer[2]+1)/2.0f;
    //slika[i + kamera.sirina * j] = RGB{tmp2, tmp2, tmp2};
    slika[i + kamera.sirina * (kamera.visina - 1 - j)] = piksel;
      }
}

/*__global__ void SlikaGPU::transformKernel(ScenaPodaci scena)
{
  int i = hipThreadIdx_x;
  if(i < scena.brojSfera)
    scena.sfere[i].Transformiraj(0);

  else
    scena.trokuti[i - scena.brojSfera].Transformiraj(0);
    }
*/

__global__ void SlikaGPU::transformKernel(ScenaPodaci scena, size_t offset, size_t brojSferaNiti, size_t brojTrokutNiti, size_t brojModelNiti)
{
  size_t i = hipBlockDim_x * hipBlockIdx_x + hipThreadIdx_x + offset;
  if(i >= brojSferaNiti + brojTrokutNiti + brojModelNiti + offset)
    {
      if(i < scena.brojSvjetala)
	scena.svjetla[i].Transformiraj();
      else
	return;
    }

  if(i < scena.brojSvjetala)//
    scena.svjetla[i].Transformiraj(); 
  
  if(i < brojSferaNiti)
    scena.sfere[i].Transformiraj(0);

  else if(i < brojTrokutNiti + brojSferaNiti)
    scena.trokuti[i - brojSferaNiti].Transformiraj(0);

  else
    { 
      size_t nitiDoSada = brojSferaNiti + brojTrokutNiti + offset;
      for(int j = 0; j < scena.brojTrokutModela; ++j)
	{
	  TrokutModelGPU model(scena.modeli[j]);
	  
	  if(i >= nitiDoSada &&
	     i < nitiDoSada + model.DajBrojNitiZaTransform())
	    {
	      model.Transformiraj(i - nitiDoSada);
	      break;
	    }

	  nitiDoSada += model.DajBrojNitiZaTransform();	     
	}
    }
}

__global__ void SlikaGPU::rayKernel(ScenaPodaci scena, Kamera kamera, Uint32 *slika)
{
  int i = hipBlockIdx_x * hipBlockDim_x + hipThreadIdx_x;
  int j = hipBlockIdx_y * hipBlockDim_y + hipThreadIdx_y;

  if(i >= kamera.sirina || j >= kamera.visina)
    return; //izašli smo van ekrana; odustani od rendera za ovaj piksel!
  //probaj optimizaciju: ovo možda dovodi do prevelike divergencije warpova,
  //pa bi bilo bolje alocirati veću padding rezoluciju i nju renderirati
  //na GPU-u, bez obzira što CPU neće iskoristiti cijelu tu površinu kad
  //pročita sliku (orig.rezolucija je ovdje i ona se vraća sa getterima
  //, dok interno u privatnim varijablama GPU cacheira virtualnu rez.!)
  
  Realan u = kamera.l + i + 0.5;
  Realan v = kamera.b + j + 0.5;

  Vektor3 smjer, o;
  switch(kamera.projekcija)//optimizacija: 2 vrste kernela, za 2 vrste kamere jer jedna ne može odjednom postati druga dinamički i ovdje je to suvišno granati!
    {
    case Kamera::TipProjekcije::Ortografska:
      o = kamera.pozicija.Radijusvektor() + u * kamera.u + v * kamera.v;
      smjer = -kamera.w;
      break;

    case Kamera::TipProjekcije::Perspektivna:
      o = kamera.pozicija.Radijusvektor();
      smjer = -kamera.d * kamera.w + u * kamera.u + v * kamera.v;
      break;
    }

  Zraka zraka(o, smjer);
  SferaGPU *prvaSfera = nullptr;
  TrokutModelBaza::Podaci *prviKvadar = nullptr;
  const TrokutGPU *prviTrokut = nullptr;
  Realan najmanji = BESKONACNO;
  Vektor3 potencijalnoSjeciste, sjeciste;
  int presjecenTrokut;

  Realan t = BESKONACNO;
  for(int k = 0; k < scena.brojTrokutModela; ++k)
    {
      TrokutModelGPU objekt(scena.modeli[k]);
      //Realan t = BESKONACNO;
      if (!(objekt.PresijeciZraku(zraka, t, presjecenTrokut, potencijalnoSjeciste)))
	continue;

      if (t < najmanji && t > 0)//iza svih i ispred kamere
	{
	  najmanji = t;
	  prviKvadar = &scena.modeli[k];
	  prvaSfera = nullptr;
	  prviTrokut = nullptr;
	  sjeciste = potencijalnoSjeciste;
	}
    }

    for(int k = 0; k < scena.brojTrokuta; ++k)
      {
      const auto &objekt = scena.trokuti[k];
      Realan t;
      if (!(objekt.PresijeciZraku(zraka, t, potencijalnoSjeciste)))
	continue;
      
      if (t < najmanji && t > 0)//iza svih i ispred kamere
	{
	  najmanji = t;
	  prviTrokut = &objekt;
	  prvaSfera = nullptr;
	  prviKvadar = nullptr;
	  sjeciste = potencijalnoSjeciste;
	}
      }

    for(int k = 0; k < scena.brojSfera; ++k) //TODO:prebaci ovo generično računanje u template funkciju (izgubismo polimorfizam)
    {
      auto &objekt = scena.sfere[k];
      Realan t;
      if (!(objekt.PresijeciZraku(zraka, t, potencijalnoSjeciste)))
	continue;

      if (t < najmanji && t > 0)//iza svih i ispred kamere
	{
	  najmanji = t;
	  prvaSfera = &objekt;
	  prviKvadar = nullptr;
	  prviTrokut = nullptr;
	  sjeciste = potencijalnoSjeciste;
	}
    }


    Uint32 piksel;//RGB piksel;
    if(!prviTrokut && !prviKvadar && !prvaSfera)
      {
	smjer = smjer.Normiraj();
	Realan tmp0 = (smjer[0]+1)/2.0f;
	Realan tmp1 = (smjer[1]+1)/2.0f;
	Realan tmp2 = (smjer[2]+1)/2.0f;
	//piksel = pozadinskaBoja;
	piksel = RGB{tmp1*0.4f, tmp1*0.5f, tmp1};
      }

    else
      {
	RGB rgb;
	if(prviTrokut)
	  {
	    //rgb = RGB{1.0f, 1.0f, 1.0f};
	    rgb = prviTrokut->Sjencaj(sjeciste, scena.svjetla, scena.brojSvjetala);
	  }

	else if(prviKvadar)
	  {
	    rgb = RGB{1.0f, 0.0f, 0.0f};
	    
	    TrokutModelGPU model(*prviKvadar);
	    rgb = model.Sjencaj(presjecenTrokut, sjeciste, scena.svjetla, scena.brojSvjetala);
	    
	  }

	else if(prvaSfera)
	  {
	    rgb = prvaSfera->Sjencaj(sjeciste, scena.svjetla, scena.brojSvjetala);
	  }

	piksel = rgb;
	//piksel = RGB{255u, 255u, 255u};
      }

    //    Realan tmpx = static_cast<Realan>(i)/kamera.sirina;
    //Realan tmpy = static_cast<Realan>(j)/kamera.visina;
    //slika[i + kamera.sirina * j] = RGB{tmpx, tmpx, tmpx} * RGB{tmpy, 0.0f, 0.0f};

    //slika[i + kamera.sirina * j] = RGB{tmp2, tmp2, tmp2};
    slika[i + kamera.sirina * (kamera.visina - 1 - j)] = piksel;
}

SlikaGPU &SlikaGPU::operator=(const SlikaGPU &slikaGPU) //TODO:ako se već ovako koristi =, onda bi copy cons isto trebao raditi deep copy, a Uint32 ptr fb-a bi se trebao slati kernelu budući da nema smisla sve kopirati, već je na GPU-u tada
{
  if(slika)
    hipFree(slika);

  auto dimenzije = slikaGPU.DajDimenzije();
  sirina = dimenzije.first;
  visina = dimenzije.second;
  auto brojBlokova = slikaGPU.DajBrojBlokova();
  brojBlokX = brojBlokova.first;
  brojBlokY = brojBlokova.second;
  
  hipMalloc(&slika, sizeof(Uint32) * sirina * visina);
  return *this;
}

SlikaGPU &SlikaGPU::operator=(SlikaGPU &&slika2)
{
  //slika = std::exchange(slika2.DajSliku(), slika); //HIP nema std::exchange...
  Uint32 *tmp = slika;
  slika = slika2.DajSliku();
  slika2.PostaviSliku(tmp);
    
  sirina = slika2.DajDimenzije().first;
  visina = slika2.DajDimenzije().second;
  
  brojBlokX = slika2.DajBrojBlokova().first;
  brojBlokY = slika2.DajBrojBlokova().second;
  return *this;
}

__host__ void SlikaGPU::TransformirajSvjetla(ScenaPodaci &scena)
{
  for(size_t i = 0; i < scena.brojSvjetala; ++i)
    scena.svjetla[i].Transformiraj();
}

__host__ void SlikaGPU::Raytrace(ScenaGPU &scena, Kamera &kamera)
{
  int brojObjekata = scena.DajBrojSfera() + scena.DajBrojTrokuta();
  //  TransformirajSvjetla(scena.DajPodatke());
  
  // hipLaunchKernelGGL(SlikaGPU::transformKernel, dim3(1), dim3(brojObjekata), 0, 0, scena.DajPodatke());
  
  int blokova = (scena.DajBrojTransformNiti() + scena.DajBrojSvjetala() + MAX_NITI_PO_BLOKU - 1) / MAX_NITI_PO_BLOKU; //tu zbrajamo i broj svjetala kako bi bili sigurni da su sva svjetla sigurno transformirana; to je inače istina jer svjetala ima tipično manje od ostalih objekata, ali za slučaj abnormalnosti imamo ovaj općenit slučaj
  int iteracija = (blokova + MAX_BLOKOVA - 1) / MAX_BLOKOVA; //transform iteracija
  size_t brojNiti = MAX_NITI_PO_BLOKU;
  size_t ukupnoNiti = 0;
  int ostatak = blokova % MAX_BLOKOVA;

  for(int i = 0; i < iteracija; ++i)
    {
      if(i == iteracija-1 && ostatak != 0)
	{
	  //brojNiti = ostatak * MAX_NITI_PO_BLOKU;
	  blokova = ostatak;
	}
      hipLaunchKernelGGL(SlikaGPU::transformKernel, dim3(blokova),
			 dim3(brojNiti), 0, 0, scena.DajPodatke(),
			 ukupnoNiti,
			 scena.DajBrojSferaTransformNiti(),
			 scena.DajBrojTrokutTransformNiti(),
			 scena.DajBrojModelTransformNiti());

      ukupnoNiti += blokova * brojNiti;
  
    }
  
  hipLaunchKernelGGL(SlikaGPU::rayKernel, dim3(brojBlokX, brojBlokY), dim3(BLOK_X, BLOK_Y), 0, 0, scena.DajPodatke(), kamera, slika);
}

__host__ void SlikaGPU::RaytraceCPU(ScenaGPU &scena, Kamera &kamera, Uint32 *slika2)
{
  ScenaPodaci podaci;
  podaci.brojSvjetala = scena.DajPodatke().brojSvjetala;
  podaci.brojSfera = scena.DajPodatke().brojSfera;
  podaci.brojTrokuta = scena.DajPodatke().brojTrokuta;

  podaci.svjetla = new Svjetlo[podaci.brojSvjetala * sizeof(Svjetlo)];
  podaci.sfere = new SferaGPU[podaci.brojSfera * sizeof(SferaGPU)];
  podaci.trokuti = new TrokutGPU[podaci.brojTrokuta * sizeof(TrokutGPU)];

  hipMemcpy(podaci.svjetla, scena.DajPodatke().svjetla, sizeof(Svjetlo) * podaci.brojSvjetala, hipMemcpyDeviceToHost);
  hipMemcpy(podaci.sfere, scena.DajPodatke().sfere, sizeof(SferaGPU) * podaci.brojSfera, hipMemcpyDeviceToHost);
  hipMemcpy(podaci.trokuti, scena.DajPodatke().trokuti, sizeof(TrokutGPU) * podaci.brojTrokuta, hipMemcpyDeviceToHost);
  ScenaGPU::TransformirajObjekte(podaci);
  rayCPU(podaci, kamera, slika2);
}
