#pragma once
//indeksirani trokutasti mesh
#include "ObjektGPU.h"
#include "Vektor3.h"
#include "TrokutModelBaza.h"
#include "TrokutModelGPU.h"
#include "SjeceTrokut.h"
#include "SjencaTrokut.h"
#include "GPUCopyable.h"
#include "OBJ.h"
#include <vector>

class TrokutModelGPU;
class Scena;

/** Tip složenijeg indeksiranog trokutastog mesha na CPU-u.
 *
 *  Ovaj tip objekta zahtijeva dodatne alokacije i memorijski menadžment pa se za sada njegovo renderiranje radi odvojeno za GPU i za CPU. Ova klasa radi CPU stranu na kojoj svaki model počinje, a tada po potrebi korisnika može biti prekopiran na GPU i tamo renderiran preko ScenaGPU.
 */
class TrokutModel final : public TrokutModelBaza, public GPUCopyable<TrokutModel>
{

 public:
  friend Scena;
  
  __host__ TrokutModel() = default;
  //  __host__ TrokutModel(const TrokutModelBaza::Podaci &novi){podaci = novi;}
  /** Konstruktor modela na temelju OBJ datoteke.
   *  \param[in] obj   Jedan učitan objekt iz OBJ datoteke.
   */
  __host__ TrokutModel(const ModelOBJ::Objekt &obj);
  //još neki konstruktori...?
  __host__ TrokutModel(const TrokutModel &);
  __host__ TrokutModel(TrokutModel&&) = default;
  __host__ TrokutModel &operator=(const TrokutModel&);
  __host__ TrokutModel &operator=(TrokutModel&&);
  __host__ ~TrokutModel();

  /** CPU-specifičan shader za trokutasti mesh.
   *  \param[in] sjeciste  Izračunato sjecište.
   *  \param[in] svjetla   Lista svih svjetala na sceni.
   *  \return RGB boja piksela.
   */
  __host__ RGB Sjencaj(int,const Vektor3 &sjeciste, const std::forward_list<Svjetlo> &svjetla);
  /** Siječe zraku i cijeli mesh.
   *
   *  Ovo može biti vrlo skupo i zato treba ubrzavati s različitim hijerarhijskim akcelaracijskim strukturama. Za sada se jednostavno testiraju svi trokuti uz neke manje interne optimizacije. Za razliku od ostalih metoda za presjecanje, ova vraća i indeks presječenog trokuta koji je potreban radi sjenčanja kasnije.
   *
   * Ovo je CPU verzija.
   * \param[in] zraka   Zraka koju provjeravamo.
   * \param[out] t      Parametar \f$t\f$ u jednadžbi zrake koji vodi do sjecišta.
   * \param[out] sjeciste Izračunato sjecište unutar trokuta u *baricentričnim koordinatama*!
   * \return `true` ako zraka siječe bilo koji trokut mesha; inače `false`.
   */
  __host__ bool PresijeciZraku(const Zraka &zraka,Realan &t,int&,Vektor3 &sjeciste);

  ///  TrokutModelGPU PrekopirajNaGPU() const override;
  /** Prekopira model u GPU memoriju.
   *  \return Konstruirana kopija tipa TrokutModelGPU na GPU-u.
   */
  TrokutModelGPU PrekopirajNaGPU() override;

  /** Primijeni zadanu transformacijsku matricu na cijeli mesh.
   *  \param[in] mat   Zadana matrica.
   */
  __host__ void PrimijeniTransformaciju(const Matrica4x4 &mat);
  __host__ void Transformiraj(int) override;

  __device__ __host__ Vektor3 *DajVrhove() const {return podaci.vrhovi;}
  __device__ __host__ size_t DajBrojVrhova() const {return podaci.brojVrhova;}
  __device__ __host__ Vektor3 *DajNormale() const {return podaci.normale;}
  __device__ __host__ size_t DajBrojNormala() const {return podaci.brojNormala;}
  __device__ __host__ Vektor3 *DajUV() const {return podaci.UV;}
  __device__ __host__ size_t DajBrojUV() const {return podaci.brojUV;}
  __device__ __host__ size_t DajBrojPoligona() const {return podaci.brojPoligona;}
  __device__ __host__ Poligon *DajPoligone() const {return podaci.poligoni;}
  __host__ void PostaviVrhove(Vektor3 *nv) {podaci.vrhovi = nv;}
  __host__ void PostaviNormale(Vektor3 *nn) {podaci.normale = nn;}
  __host__ void PostaviUV(Vektor3 *nt) {podaci.UV = nt;}
  __host__ void PostaviPoligone(Poligon *p) {podaci.poligoni = p;}


 private:
  /** Pomoćna funkcija za korekciju indeksa raznih elemenata modela.
   *
   *  Ova funkcija služi određivanju konačnog, apsolutnog i 0-baziranog indeksa danog elementa iz OBJ formata, a koji se tiče neke konkretne grupe vrhova ili normala koji inače mogu u tom formatu biti specificirati relativno (stoga nam treba njihova ukupna količina) i originalno su 1-bazirani.
   *  \param[in] idx    Originalni indeks koji treba korigirati.
   *  \param[in] kolicina Ukupna količina danog elementa (npr. vrhova, normala modela i sl.).
   *  \return Točan apsolutni indeks u odgovarajuće učitano polje elemenata.
   */
  __host__ int SrediIndeks(int idx, int kolicina);
  /** Pomoćna funkcija za kopiranje parsiranih OBJ podataka u internu strukturu s dodatnim optimizacijama i računima po potrebi.
   *  \param[in] obj   Parsirani OBJ model/objekt.
   */
  __host__ void Kopiraj(const ModelOBJ::Objekt &obj);
  /** Oslobađa svu memoriju vezanu za model iz host prostora.
   */
  __host__ void Unisti();
  
 protected:
  __host__ void Dealociraj();
  __host__ void Realociraj(const TrokutModel&);

};


