/*
Copyright Mauro Raguzin GPLv3 2020
*/
#include "MatBaza.h"
#include "Scena.h"
#include "ScenaGPU.h"
#include "SferaGPU.h"
#include "TrokutGPU.h"
#include "TrokutModel.h"
#include "Kamera.h"
#include "Svjetlo.h"
#include "Slika.h"
#include "SlikaGPU.h"
#include "SDLIntegracija.h"
#include "Prozor.h"
#include "OBJ.h"
#include <iostream>
#include <cstdlib>
#include <chrono>

/** \file main_hip.cpp
 *  Demo aplikacija raytracera.
 */

struct Postavke
{
  int w = 640, h = 480;
  Realan fov = 90;
  std::vector<std::string> modelFajlovi;
  bool cpuRay = false;
};

/** Pomoćna funkcija za procesiranje komandne linije.
 *
 *  Opcije komandne linije
 *  =======================
 *  Opcija   | Učinak
 *  -------- | ---------
 *  -cpu     | Renderiraj scenu isključivo na CPU-u
 *  -w <broj>| Specificira širinu prozora u pikselima
 *  -h <broj>| Specificira visinu prozora u pikselima
 *  *.obj    | Specificira .obj trokutasti mesh model; više .obj-ova je moguće i u tom slučaju se slažu horiontalno u liniji po sceni
 */
Postavke SrediArgumente(int argc, char *argv[])
{
  Postavke postavke;
  bool citajW = false;
  bool citajH = false;
  
  for(int i = 1; i < argc; ++i)
    {
      std::string arg(argv[i]);

      if(citajW)
	{
	  postavke.w = atoi(argv[i]);
	  citajW = false;
	}

      else if(citajH)
	{
	  postavke.h = atoi(argv[i]);
	  citajH = false;
	}
      
      else if(arg == "-cpu")
	postavke.cpuRay = true;

      else if(arg.find(".obj") == arg.length()-4) //pretpostavljamo lowercase!
	postavke.modelFajlovi.push_back(arg);

      else if(arg == "-w")
	citajW = true;

      else if(arg == "-h")
	citajH = true;
    }

  return postavke;
}

/** Glavni kod za iscrtavanje svakog framea.
 */
bool Iteracija(Scena &scenaCPU, ScenaGPU &scena, SlikaGPU &slikaGPU, Slika &slika, Prozor &prozor, Kamera &kamera, bool cpuRay)
{
  if(!prozor.DajProzor())
    return false;

  auto start = std::chrono::steady_clock::now();
  
  kamera.GledajU(Vektor4{0, 0, -3, 1.f}, Vektor3{0.0f, 1.0f, 0.0f});
  if(!cpuRay)
    {
      slikaGPU.Raytrace(scena, kamera);
      //slikaGPU.RaytraceCPU(scena, kamera, slika.DajSliku());
      slika = slikaGPU;
      prozor.Vsync(slika.DajSliku());
    }

  else
    {
      slika.Raytrace(kamera, scenaCPU);
      prozor.Vsync(slika.DajSliku());
    }

  auto end = std::chrono::steady_clock::now();
  std::chrono::duration<Realan, std::milli> interval = end - start;
  std::cout << "Frametime: " << interval.count() << "ms\n";

  return true;
}

/** Povezuje event handler za pritiske tipki na tipkovnici.
 *
 *  Kontrole kamere
 *  ========
 *  Akcija                   | Kontrola (tipka)
 *  ------------------------ | -----------------
 *  Pomak ulijevo            | Lijevi kursor
 *  Pomak udesno             | Desni kursor
 *  Pomak na gore            | Page Up
 *  Pomak na dolje           | Page Down
 *  Pomak naprijed           | Kursor naprijed
 *  Pomak unazad             | Kursor unazad
 *  Zumiranje                | +
 *  Odzumiravanje            | -
 *  Promjena tipa projekcije | Enter
 */
void PoveziTipke(Kamera &kamera, Prozor &prozor)
{
  prozor.PoveziTipke([&](SDL_Keycode tipka)
		     {
		       auto pozicija = kamera.Pozicija();
		       Realan hfov = kamera.HFOV();
		       auto pomak = Matrica4x4::Jedinicna();
		       auto tipKamere = kamera.Tip();
		       
		       switch(tipka)
			 {
			 case SDLK_PLUS:
			   hfov -= 0.2;//SMANJI fov
			   kamera.PostaviFOV(hfov);
			   break;

			 case SDLK_MINUS:
			   hfov += 0.2;
			   kamera.PostaviFOV(hfov);
			   break;

			 case SDLK_RETURN:
			   if(tipKamere == Kamera::TipProjekcije::Perspektivna)
			     tipKamere = Kamera::TipProjekcije::Ortografska;
			   else
			     tipKamere = Kamera::TipProjekcije::Perspektivna;

			   kamera.PostaviTip(tipKamere);
			   break;

			 case SDLK_UP:
			   pomak = Matrica4x4::Translatiraj(-10.0f * kamera.W());
			   kamera.KomponirajTransformaciju(pomak);
			   break;

			 case SDLK_DOWN:
			   pomak = Matrica4x4::Translatiraj(10.0f * kamera.W());
			   kamera.KomponirajTransformaciju(pomak);
			   break;

			 case SDLK_PAGEUP:
			   pomak = Matrica4x4::Translatiraj(10.0f * kamera.V());
			   kamera.KomponirajTransformaciju(pomak);
			   break;

			 case SDLK_PAGEDOWN:
			   pomak = Matrica4x4::Translatiraj(-10.0f * kamera.V());
			   kamera.KomponirajTransformaciju(pomak);
			   break;

			 case SDLK_RIGHT:
			   pomak = Matrica4x4::Translatiraj(10.0f * kamera.U());
			   kamera.KomponirajTransformaciju(pomak);
			   break;

			 case SDLK_LEFT:
			   pomak = Matrica4x4::Translatiraj(-10.0f * kamera.U());
			   kamera.KomponirajTransformaciju(pomak);
			   break;

			   /* TODO: treba li nam ovo?
			 case SDLK_c:
			   SDL_SetWindowSize(prozor, w - 100, h - 100);
			   SDL_Event noviEvent;
			   noviEvent.type = SDL_WINDOWEVENT;
			   noviEvent.window.event = SDL_WINDOWEVENT_RESIZED;
			   noviEvent.window.data1 = w - 100;
			   noviEvent.window.data2 = h - 100;
			   SDL_PushEvent(&noviEvent);
			   break;
			   */
			 }
		     });  
}

void PoveziHandlere(Prozor &prozor, Kamera &kamera, SlikaGPU &slika)
{
  prozor.PoveziPromjenu([&](std::pair<int,int> noveDim)
			{
			  slika = SlikaGPU(noveDim.first, noveDim.second);
			  auto pozicija = kamera.Pozicija();
			  Realan fov = kamera.HFOV();
			  //kamera = Kamera(Kamera::TipProjekcije::Perspektivna, Vektor4(pozicija[0], pozicija[1], pozicija[2], 1), Vektor3(1, 0, 0), Vektor3(0, 1, 0), Vektor3(0, 0, 1), MatBaza::Radijana(fov), noveDim.first, noveDim.second);
			  kamera.PostaviDimenzije(noveDim.first, noveDim.second);
			  std::cout << "Promjena rezolucije na (" << noveDim.first << "," << noveDim.second << ")\n";
			});

  PoveziTipke(kamera, prozor);

}


int main(int argc, char *argv[])
{
  Postavke postavke = SrediArgumente(argc, argv);
  SlikaGPU slikaGPU(postavke.w, postavke.h);
  Scena scena;
  auto start = std::chrono::steady_clock::now();

  Vektor3 trans(60, 0, 0);
  for(auto &ime : postavke.modelFajlovi)
    {
      ModelOBJ obj(ime);
      TrokutModel model(obj.DajObjekte()[0]);
      model.PrimijeniTransformaciju(Matrica4x4::Translatiraj(trans)
				    * Matrica4x4::Skaliraj(20));
      
      scena.DodajModel(std::move(model));

      std::cout << trans[0] << '\n';
      trans += Vektor3(-50, 5, 0);
    }
  
  Kamera kamera(Kamera::TipProjekcije::Perspektivna, {0.0f, 0.0f, 0.0f, 1.0f},
		{1.0f, 0.0f, 0.0f}, {0.0f, 1.0f, 0.0f}, {0.0f, 0.0f, 1.0f},
		postavke.fov, postavke.w, postavke.h);
		

  TrokutGPU trokut1(Vektor4{0.0f, 0.0f, 0.0f, 1.0f},
		    Vektor4{0.0f, 200.0f, 0.0f, 1.0f},
		    Vektor4{0.0f, 0.0f, -200.0f, 1.0f}, RGB{255u, 255u, 255u});
  trokut1.transformacija = Matrica4x4::Translatiraj({0, 30, -10}) *
    Matrica4x4::RotirajY(MatBaza::Radijana(80));

  SferaGPU sfera1(10.0f, {255u, 128u, 100u});
  sfera1.transformacija = Matrica4x4::Translatiraj({-10.0f, 5.0f, 15.0f});
  SferaGPU sfera2(5.0f, {255u, 128u, 100u});
  sfera2.transformacija = Matrica4x4::Translatiraj({5.0f, 5.0f, -3.0f});
  SferaGPU sfera3(7.0f, {255u, 0u, 0u});
  sfera2.transformacija = Matrica4x4::Translatiraj({0.0f, -20.0f, 0.0f});

  // SPONZA
  //kocka.transformacija = Matrica4x4(100, 0, 0, 0,
  //					 0, 100, 0, 0,
  //					 0,  0, 100,0,
  //				    0,  0,  0, 1) * Matrica4x4::Translatiraj({0, 160, 0});
  
  //std::cout << "kocka mat[0][0]=" << kocka.podaci.transformacija.Element(0,0)<<'\n';
  //  kvadar1.transformacija = Matrica4x4::Translatiraj({0.0f, -8.0f, -8.0f});
  Svjetlo svjetlo1(Vektor3(-5, 50, -500),{128u,255u,65u});
  Svjetlo svjetlo0(Vektor3(), {255u, 255u, 255u});
  //Svjetlo svjetlo1(Vektor3{0.0f, 0.0f, 0.0f}, {255u, 255u, 255u});

  scena.DodajTrokut(trokut1);
  scena.DodajSferu(sfera1);
  scena.DodajSferu(sfera2);
  scena.DodajSferu(sfera3);
  scena.DodajSvjetlo(svjetlo1);
  scena.DodajSvjetlo(svjetlo0);

  auto &sdl = SDLIntegracija::DajInstancu(SDL_INIT_VIDEO);
  Prozor prozor(sdl, "Projektni zadatak: GPU raytracer (Mauro Raguzin)", postavke.w, postavke.h);
  PoveziHandlere(prozor, kamera, slikaGPU);

  std::cout << "objekti učitani\n";
  ScenaGPU scenaGPU(scena);
  std::cout << "Broj trokuta: " << scenaGPU.DajBrojTrokuta() << std::endl;
  std::cout << "total niti: " << scenaGPU.DajBrojTransformNiti() << std::endl;
  ///Slika slika(slikaGPU);
  Slika slika(postavke.w, postavke.h);

  bool dalje = true;
  size_t frames = 0;
  while(dalje)
    {
      dalje = Iteracija(scena, scenaGPU, slikaGPU, slika, prozor, kamera, postavke.cpuRay);
      ++frames;
    }

  auto end = std::chrono::steady_clock::now();
  std::chrono::duration<Realan> interval = end - start;
  std::cout << "===============\nRenderirano " << frames << " frameova u "
	    << interval.count() << " sekundi" << std::endl;
  std::cout << "Prosječni fps: " <<
    static_cast<Realan>(frames)/interval.count() << std::endl;
  return 0;
}
		    
